# -*- coding: utf-8 -*-
#
'''
I/O for h5m, cf. <https://trac.mcs.anl.gov/projects/ITAPS/wiki/MOAB/h5m>.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>
'''
from datetime import datetime
import logging
import numpy

from . import __about__


# def _int_to_bool_list(num):
#     # From <https://stackoverflow.com/a/33608387/353337>.
#     bin_string = format(num, '04b')
#     return [x == '1' for x in bin_string[::-1]]


def read(filename):
    '''Reads H5M files, cf.
    https://trac.mcs.anl.gov/projects/ITAPS/wiki/MOAB/h5m.
    '''
    import h5py

    f = h5py.File(filename, 'r')
    dset = f['tstt']

    points = dset['nodes']['coordinates'][()]
    # read point data
    point_data = {}
    if 'tags' in dset['nodes']:
        for name, dataset in dset['nodes']['tags'].items():
            point_data[name] = dataset[()]

    # # Assert that the GLOBAL_IDs are contiguous.
    # point_gids = dset['nodes']['tags']['GLOBAL_ID'][()]
    # point_start_gid = dset['nodes']['coordinates'].attrs['start_id']
    # point_end_gid = point_start_gid + len(point_gids) - 1
    # assert all(point_gids == range(point_start_gid, point_end_gid + 1))

    h5m_to_meshio_type = {
        'Edge2': 'line',
        'Tri3': 'triangle',
        'Tet4': 'tetra'
        }
    cells = {}
    cell_data = {}
    for h5m_type, data in dset['elements'].items():
        meshio_type = h5m_to_meshio_type[h5m_type]
        conn = data['connectivity']
        # Note that the indices are off by 1 in h5m.
        cells[meshio_type] = conn[()] - 1

        # TODO bring cell data back
        # if 'tags' in data:
        #     for name, dataset in data['tags'].items():
        #         cell_data[name] = dataset[()]

    # The `sets` in H5M are special in that they represent a segration of data
    # in the current file, particularly by a load balancer (Metis, Zoltan,
    # etc.). This segregation has no equivalent in other data types, but is
    # certainly worthwhile visualizing.
    # Hence, we will translate the sets into cell data with the prefix "set::"
    # here.
    field_data = {}
    # TODO deal with sets
    # if 'sets' in dset and 'contents' in dset['sets']:
    #     # read sets
    #     sets_contents = dset['sets']['contents'][()]
    #     sets_list = dset['sets']['list'][()]
    #     sets_tags = dset['sets']['tags']

    #     cell_start_gid = conn.attrs['start_id']
    #     cell_gids = cell_start_gid + elems['tags']['GLOBAL_ID'][()]
    #     cell_end_gid = cell_start_gid + len(cell_gids) - 1
    #     assert all(cell_gids == range(cell_start_gid, cell_end_gid + 1))

    #     # create the sets
    #     for key, value in sets_tags.items():
    #         mod_key = 'set::' + key
    #         cell_data[mod_key] = numpy.empty(len(cells), dtype=int)
    #         end = 0
    #         for k, row in enumerate(sets_list):
    #             bits = _int_to_bool_list(row[3])
    #             # is_owner = bits[0]
    #             # is_unique = bits[1]
    #             # is_ordered = bits[2]
    #             is_range_compressed = bits[3]
    #             if is_range_compressed:
    #                 start_gids = sets_contents[end:row[0]+1:2]
    #                 lengths = sets_contents[end+1:row[0]+1:2]
    #                 for start_gid, length in zip(start_gids, lengths):
    #                     end_gid = start_gid + length - 1
    #                     if start_gid >= cell_start_gid and \
    #                             end_gid <= cell_end_gid:
    #                         i0 = start_gid - cell_start_gid
    #                         i1 = end_gid - cell_start_gid + 1
    #                         cell_data[mod_key][i0:i1] = value[k]
    #                     else:
    #                         # TODO deal with point data
    #                         raise RuntimeError('')
    #             else:
    #                 gids = sets_contents[end:row[0]+1]
    #                 cell_data[mod_key][gids - cell_start_gid] = value[k]

    #             end = row[0] + 1

    return points, cells, point_data, cell_data, field_data


def write(filename,
          points,
          cells,
          point_data=None,
          cell_data=None,
          field_data=None,
          add_global_ids=True):
    '''Writes H5M files, cf.
    https://trac.mcs.anl.gov/projects/ITAPS/wiki/MOAB/h5m.
    '''
    import h5py

    point_data = {} if point_data is None else point_data
    cell_data = {} if cell_data is None else cell_data
    field_data = {} if field_data is None else field_data

    f = h5py.File(filename, 'w')

    tstt = f.create_group('tstt')

    # The base index for h5m is 1.
    global_id = 1

    # add nodes
    nodes = tstt.create_group('nodes')
    coords = nodes.create_dataset('coordinates', data=points)
    coords.attrs.create('start_id', global_id)
    global_id += len(points)

    # Global tags
    tstt_tags = tstt.create_group('tags')

    # The GLOBAL_ID associated with a point is used to identify points if
    # distributed across several processes. mbpart automatically adds them,
    # too.
    if 'GLOBAL_ID' not in point_data and add_global_ids:
        point_data['GLOBAL_ID'] = numpy.arange(1, len(points)+1, )

    # add point data
    if point_data is not None:
        tags = nodes.create_group('tags')
        for key, data in point_data.items():
            if len(data.shape) == 1:
                dtype = data.dtype
                tags.create_dataset(key, data=data)
            else:
                # H5M doesn't accept n-x-k arrays as data; it wants an n-x-1
                # array with k-tuples as entries.
                n, k = data.shape
                dtype = numpy.dtype((data.dtype, (k,)))
                dset = tags.create_dataset(key, (n,), dtype=dtype)
                dset[:] = data

            # Create entry in global tags
            g = tstt_tags.create_group(key)
            g['type'] = dtype
            # Add a class tag:
            # From
            # <http://lists.mcs.anl.gov/pipermail/moab-dev/2015/007104.html>:
            # ```
            # /* Was dense tag data in mesh database */
            #  define mhdf_DENSE_TYPE   2
            # /** \brief Was sparse tag data in mesh database */
            # #define mhdf_SPARSE_TYPE  1
            # /** \brief Was bit-field tag data in mesh database */
            # #define mhdf_BIT_TYPE     0
            # /** \brief Unused */
            # #define mhdf_MESH_TYPE    3
            #
            g.attrs['class'] = 2

    # add elements
    elements = tstt.create_group('elements')

    elem_dt = h5py.special_dtype(
        enum=('i', {
            'Edge': 1,
            'Tri': 2,
            'Quad': 3,
            'Polygon': 4,
            'Tet': 5,
            'Pyramid': 6,
            'Prism': 7,
            'Knife': 8,
            'Hex': 9,
            'Polyhedron': 10
            })
        )

    tstt['elemtypes'] = elem_dt

    tstt.create_dataset(
        'history',
        data=[
            __name__.encode('utf-8'),
            __about__.__version__.encode('utf-8'),
            str(datetime.now()).encode('utf-8')
            ])

    # number of nodes to h5m name, element type
    meshio_to_h5m_type = {
        'line': {'name': 'Edge2', 'type': 1},
        'triangle': {'name': 'Tri3', 'type': 2},
        'tetra': {'name': 'Tet4', 'type': 5}
        }
    for key, data in cells.items():
        if key not in meshio_to_h5m_type:
            logging.warning(
                'Unsupported H5M element type \'%s\'. Skipping.', key
                )
            continue
        this_type = meshio_to_h5m_type[key]
        elem_group = elements.create_group(this_type['name'])
        elem_group.attrs.create(
            'element_type',
            this_type['type'],
            dtype=elem_dt
            )
        # h5m node indices are 1-based
        conn = elem_group.create_dataset('connectivity', data=(data + 1))
        conn.attrs.create('start_id', global_id)
        global_id += len(data)

    # add cell data
    if cell_data:
        tags = elem_group.create_group('tags')
        for key, value in cell_data.items():
            tags.create_dataset(key, data=value)

    # add empty set -- MOAB wants this
    sets = tstt.create_group('sets')
    sets.create_group('tags')

    # set max_id
    tstt.attrs.create('max_id', global_id, dtype='u8')
    return



# -*- coding: utf-8 -*-
#
from __future__ import print_function

from .__about__ import (
    __version__,
    __author__,
    __author_email__,
    __website__
    )

# pylint: disable=wildcard-import
from .helpers import *

try:
    import pipdate
except ImportError:
    pass
else:
    if pipdate.needs_checking(__name__):
        print(pipdate.check(__name__, __version__))



# -*- coding: utf-8 -*-
#
'''
I/O for DOLFIN's XML format, cf.
<https://people.sc.fsu.edu/~jburkardt/data/dolfin_xml/dolfin_xml.html>.

.. moduleauthor:: Nico Schlömer <nico.schloemer@gmail.com>
'''
import logging
import os
import re
import xml.etree.cElementTree as ET

import numpy


def _read_mesh(filename):
    dolfin_to_meshio_type = {
        'triangle': ('triangle', 3),
        'tetrahedron': ('tetra', 4),
        }

    # Use iterparse() to avoid loading the entire file via parse(). iterparse()
    # allows to discard elements (via clear()) after they have been processed.
    # See <https://stackoverflow.com/a/326541/353337>.
    for event, elem in ET.iterparse(filename, events=('start', 'end')):
        if event == 'end':
            continue

        if elem.tag == 'dolfin':
            # Don't be too strict with the assertion. Some meshe files don't
            # have the proper tags.
            # assert elem.attrib['nsmap'] \
            #     == '{\'dolfin\': \'https://fenicsproject.org/\'}'
            pass
        elif elem.tag == 'mesh':
            dim = int(elem.attrib['dim'])
            cell_type, npc = dolfin_to_meshio_type[elem.attrib['celltype']]
        elif elem.tag == 'vertices':
            points = numpy.empty((int(elem.attrib['size']), 3))
            keys = ['x', 'y']
            if dim == 2:
                points[:, 2] = 0.0
            else:
                assert dim == 3
                keys += ['z']
        elif elem.tag == 'vertex':
            k = int(elem.attrib['index'])
            points[k][:dim] = [float(elem.attrib[key]) for key in keys]
        elif elem.tag == 'cells':
            cells = {
                cell_type:
                numpy.empty((int(elem.attrib['size']), npc), dtype=int)
                }
        elif elem.tag in ['triangle', 'tetrahedron']:
            k = int(elem.attrib['index'])
            cells[cell_type][k] = [
                int(elem.attrib['v{}'.format(i)]) for i in range(npc)
                ]
        else:
            logging.warning('Unknown entry %s. Ignoring.', elem.tag)

        elem.clear()

    return points, cells, cell_type


def _read_cell_data(filename, cell_type):
    dolfin_type_to_numpy_type = {
        'int': numpy.dtype('int'),
        'float': numpy.dtype('float'),
        'uint': numpy.dtype('uint'),
        }

    cell_data = {cell_type: {}}
    dir_name = os.path.dirname(filename)
    if not os.path.dirname(filename):
        dir_name = os.getcwd()

    # Loop over all files in the same directory as `filename`.
    basename = os.path.splitext(os.path.basename(filename))[0]
    for f in os.listdir(dir_name):
        # Check if there are files by the name "<filename>_*.xml"; if yes,
        # extract the * pattern and make it the name of the data set.
        out = re.match('{}_([^\\.]+)\\.xml'.format(basename), f)
        if not out:
            continue
        name = out.group(1)
        tree = ET.parse(os.path.join(dir_name, f))
        root = tree.getroot()

        mesh_functions = list(root)
        assert len(mesh_functions) == 1
        mesh_function = mesh_functions[0]

        assert mesh_function.tag == 'mesh_function'
        size = int(mesh_function.attrib['size'])
        dtype = dolfin_type_to_numpy_type[mesh_function.attrib['type']]
        data = numpy.empty(size, dtype=dtype)
        for child in mesh_function:
            assert child.tag == 'entity'
            idx = int(child.attrib['index'])
            data[idx] = child.attrib['value']

        cell_data[cell_type][name] = data
    return cell_data


def read(filename):
    points, cells, cell_type = _read_mesh(filename)
    point_data = {}
    cell_data = _read_cell_data(filename, cell_type)
    field_data = {}
    return points, cells, point_data, cell_data, field_data


def _write_mesh(filename, points, cell_type, cells):
    stripped_cells = {cell_type: cells[cell_type]}

    dolfin = ET.Element(
        'dolfin',
        nsmap={'dolfin': 'https://fenicsproject.org/'}
        )

    meshio_to_dolfin_type = {
        'triangle': 'triangle',
        'tetra': 'tetrahedron',
        }

    if len(cells) > 1:
        discarded_cells = list(cells.keys())
        discarded_cells.remove(cell_type)
        logging.warning(
            'DOLFIN XML can only handle one cell type at a time. '
            'Using %s, discarding %s.',
            cell_type, ', '.join(discarded_cells)
            )

    dim = 2 if all(points[:, 2] == 0) else 3

    mesh = ET.SubElement(
        dolfin,
        'mesh',
        celltype=meshio_to_dolfin_type[cell_type],
        dim=str(dim)
        )
    vertices = ET.SubElement(mesh, 'vertices', size=str(len(points)))

    for k, point in enumerate(points):
        ET.SubElement(
            vertices,
            'vertex',
            index=str(k),
            x=repr(point[0]),
            y=repr(point[1]),
            z=repr(point[2])
            )

    num_cells = 0
    for cls in stripped_cells.values():
        num_cells += len(cls)

    xcells = ET.SubElement(mesh, 'cells', size=str(num_cells))
    idx = 0
    for ct, cls in stripped_cells.items():
        for cell in cls:
            cell_entry = ET.SubElement(
                xcells,
                meshio_to_dolfin_type[ct],
                index=str(idx)
                )
            for k, c in enumerate(cell):
                cell_entry.attrib['v{}'.format(k)] = str(c)
            idx += 1

    tree = ET.ElementTree(dolfin)
    tree.write(filename)
    return


def _numpy_type_to_dolfin_type(dtype):
    types = {
        'int': [numpy.int8, numpy.int16, numpy.int32, numpy.int64],
        'uint': [numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64],
        'float': [numpy.float16, numpy.float32, numpy.float64],
        }
    for key, numpy_types in types.items():
        for numpy_type in numpy_types:
            if numpy.issubdtype(dtype, numpy_type):
                return key

    assert False, 'Could not convert NumPy data type to DOLFIN data type.'
    return None



def _write_cell_data(filename, dim, cell_data):
    dolfin = ET.Element(
        'dolfin',
        nsmap={'dolfin': 'https://fenicsproject.org/'}
        )

    mesh_function = ET.SubElement(
        dolfin,
        'mesh_function',
        type=_numpy_type_to_dolfin_type(cell_data.dtype),
        dim=str(dim),
        size=str(len(cell_data))
        )

    for k, value in enumerate(cell_data):
        ET.SubElement(
            mesh_function,
            'entity',
            index=str(k),
            value=repr(value),
            )

    tree = ET.ElementTree(dolfin)
    tree.write(filename)
    return


def write(filename,
          points,
          cells,
          point_data=None,
          cell_data=None,
          field_data=None):
    logging.warning(
        'Dolfin\'s XML is a legacy format. Consider using XDMF instead.'
        )

    point_data = {} if point_data is None else point_data
    cell_data = {} if cell_data is None else cell_data
    field_data = {} if field_data is None else field_data

    if 'tetra' in cells:
        cell_type = 'tetra'
    else:
        assert 'triangle' in cells
        cell_type = 'triangle'

    _write_mesh(filename, points, cell_type, cells)

    if cell_type in cell_data:
        for key, data in cell_data[cell_type].items():
            cell_data_filename = \
                '{}_{}.xml'.format(os.path.splitext(filename)[0], key)
            dim = 2 if all(points[:, 2] == 0) else 3
            _write_cell_data(cell_data_filename, dim, numpy.array(data))
    return



# -*- coding: utf-8 -*-
#
import codecs
import os

from setuptools import setup, find_packages

# https://packaging.python.org/single_source_version/
base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, 'meshio', '__about__.py'), 'rb') as f:
    # pylint: disable=exec-used
    exec(f.read(), about)


def read(fname):
    return codecs.open(os.path.join(base_dir, fname), encoding='utf-8').read()


setup(
    name='meshio',
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    packages=find_packages(),
    description='I/O for various mesh formats',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/nschloe/meshio',
    project_urls={
        'Issues': 'https://github.com/nschloe/meshio/issues',
        },
    license=about['__license__'],
    platforms='any',
    install_requires=[
        'numpy',
        'pipdate',
        ],
    extras_require={
        'all': ['netCDF4', 'h5py'],
        'exodus': ['netCDF4'],
        'hdf5': ['h5py'],  # MED, MOAB, XDMF formats
        },
    scripts=[
        'tools/meshio-convert',
        ],
    classifiers=[
        about['__status__'],
        about['__license__'],
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering'
        ]
    )
